@@ -2,10 +2,9 @@ import matplotlib.pyplot as plt # plotting
 import os # for file/folder creations
 import keras
 from keras.models import Sequential
-from keras.layers import Dense, Dropout, Activation, Flatten
+from keras.layers import Dense, Dropout, Activation, Flatten, BatchNormalization, Input, Dense
 from keras.layers import Conv2D, MaxPooling2D
-from tensorflow.keras.models import Model
-from tensorflow.keras.layers import Input, Dense
+from keras.models import Model
 
 # =======================================================================================
 # BEG: build custom model based on args
@@ -13,6 +12,7 @@ from tensorflow.keras.layers import Input, Dense
 class TestModel:
     def __init__(self, args):
         # build model
+        self.args = args
         self.model = self.build_model(args)    
     
     def build_model(self, args):
@@ -143,31 +143,69 @@ class TestModel:
         + x is input which will be forwarded in series 
         + seq is of format
             [
-                {"out_ch": int, "z": int, 'act': 'relu'},
+                {"out_ch": int, "z": int, 'act': 'relu', 'bn': True, 'p': 0.5, 'L1': 1e-4},
                 {"out_ch": int, "z": int, 'act': 'relu', "maxpool_z": 2},
                 # -------------------------------------------------------
                 {"out_ch": int, "z": int, 'act': 'relu'},
                 ....
             ]
+        + regularize all if `args.regularize_all = L1 / L2 / L1L2` 
         """
         # bulid series
         series = []
         for config in seq:
-            # append convs and nonlin one-by-one
+            
+            # append convs and nonlin one-by-one                
+            # regularize all
+            kernel_regularizer = None
+            if ('L1' in self.args.__dict__):
+                kernel_regularizer =  keras.regularizers.l1(l1=self.args.__dict__['L1'])
+            elif ('L2' in self.args.__dict__):
+                kernel_regularizer =  keras.regularizers.l2(l2=self.args.__dict__['L2'])
+            elif ('L1L2' in self.args.__dict__):
+                kernel_regularizer =  keras.regularizers.l1_l2(
+                    l1=self.args.__dict__['L1L2'][0], 
+                    l2=self.args.__dict__['L1L2'][1]
+                )
+            # regularize >>>individually<<<
+            # (OVERWRITE `kernel_regularizer` created by regularize all)
+            if 'L1' in config:
+                kernel_regularizer =  keras.regularizers.l1(l1=config['L1'])
+            elif 'L2' in config:
+                kernel_regularizer =  keras.regularizers.l2(l2=config['L2'])
+            elif 'L1L2' in config:
+                kernel_regularizer =  keras.regularizers.l1_l2(
+                    l1=config['L1L2'][0],
+                    l2=config['L1L2'][1],
+                )
+            
             series.append(
                 self.__get_conv(
-                    out_ch   = config['out_ch'],
-                    z        = config['z'],
-                    padding  = 'same'
+                    out_ch              = config['out_ch'],
+                    z                   = config['z'],
+                    padding             = 'same',
+                    kernel_regularizer  = kernel_regularizer
                 )
             )
+            
             # activation function
             series.append(self.__get_act(config['act']))
+            
             # pool if specified
             if 'maxpool_z' in config.keys():
                 # append pool one-by-one
                 series.append(self.__get_pool(config['maxpool_z']))
-        
+            
+            # (before bn)
+            if 'p' in config.keys():
+                series.append(self.__get_dropout(config["p"]))
+                
+            # batchnorm (after dropout)
+            if 'bn' in config.keys():
+                if config['bn'] is True:
+                    # append bn one-by-one
+                    series.append(self.__get_bn())
+
         # forward through series
         """ #OVERKILL
         return Sequential(series)
@@ -181,22 +219,55 @@ class TestModel:
         + x is input
         + seq is of format
             [
-                {"out_nodes": int, "act": 'relu', 'p': 0   },
-                {"out_nodes": int, "act": 'relu', 'p': 0.5 },
+                {"out_nodes": int, "act": 'relu', 'bn': True},
+                {"out_nodes": int, "act": 'relu', 'bn': Fale, 'p': 0.5 },
                 ....
             ]
         """
         # build series
         series = []
         for config in seq:
+            
+            # regularize all
+            kernel_regularizer = None
+            if ('L1' in self.args.__dict__):
+                kernel_regularizer =  keras.regularizers.l1(l1=self.args.__dict__['L1'])
+            elif ('L2' in self.args.__dict__):
+                kernel_regularizer =  keras.regularizers.l2(l2=self.args.__dict__['L2'])
+            elif ('L1L2' in self.args.__dict__):
+                kernel_regularizer =  keras.regularizers.l1_l2(
+                    l1=self.args.__dict__['L1L2'][0], 
+                    l2=self.args.__dict__['L1L2'][1]
+                )
+            # regularize >>>individually<<<
+            # (OVERWRITE `kernel_regularizer` created by regularize all)
+            if 'L1' in config:
+                kernel_regularizer =  keras.regularizers.l1(l1=config['L1'])
+            elif 'L2' in config:
+                kernel_regularizer =  keras.regularizers.l2(l2=config['L2'])
+            elif 'L1L2' in config:
+                kernel_regularizer =  keras.regularizers.l1_l2(
+                    l1=config['L1L2'][0],
+                    l2=config['L1L2'][1],
+                )
+                
             series.append(
                 self.__get_dense(
-                    out_nodes  = config["out_nodes"], 
-                    act        = config["act"]
+                    out_nodes           = config["out_nodes"], 
+                    act                 = config["act"],
+                    kernel_regularizer  = kernel_regularizer
                 )
             )
-            # dropout
-            series.append(self.__get_dropout(config["p"]))
+            
+            # batchnorm (before dropout)
+            if 'bn' in config:
+                if config['bn'] is True:
+                    # append bn one-by-one
+                    series.append(self.__get_bn())
+                    
+            # dropout (after batchnorm)
+            if 'p' in config:
+                series.append(self.__get_dropout(config["p"]))
         
         # forward through series
         """ # OVERKILL
@@ -209,24 +280,31 @@ class TestModel:
         
     # helpers start ---------------------------------
     # general
+    def __get_bn(self):
+        return BatchNormalization()
+        
     def __get_act(self, name):
         """`name` is a string. eg. 'relu'"""
         return Activation(name)
     
     # for convs_seq
-    def __get_conv(self, out_ch, z, padding):
+    def __get_conv(self, out_ch, z, padding, kernel_regularizer):
         """
-        For "SAME" padding, if you use a stride of 1, the layer's outputs will have the same spatial dimensions as its inputs.
+        For "SAME" padding, if you use a stride of 1, 
+        the layer's outputs will have the same spatial 
+        dimensions as its inputs.
         """
-        return Conv2D(out_ch, (z, z), padding=padding, strides=(1, 1))
+        return Conv2D(out_ch, (z, z), padding=padding, 
+                      strides=(1, 1), kernel_regularizer=kernel_regularizer)
     
     def __get_pool(self, z):
         return MaxPooling2D(pool_size=(z, z))
     
     # for fcs_Seq
-    def __get_dense(self, out_nodes, act):
+    def __get_dense(self, out_nodes, act, kernel_regularizer):
         """out_nodes is int and act is string """
-        return Dense(out_nodes, activation=act)
+        print(kernel_regularizer)
+        return Dense(out_nodes, activation=act, kernel_regularizer=kernel_regularizer)
     
     def __get_dropout(self, p):
         """ p if float [0, 1]"""
@@ -237,4 +315,4 @@ class TestModel:
     # ===========================================================
 # =======================================================================================
 # END: build custom model based on args
-# =======================================================================================
\ No newline at end of file
+# =======================================================================================
